"use strict";

define([
    "jquery",
    "loglevel",
    "./icons",
    "./settings",
    "./notificationmanager",
    "./featurehandler",
    "./utils"
],
function ($, loglevel, icons, settings, notificationManager, featurehandler, utils) {
    var log,
        audioNotificationManager,
        audioInNotification = null,
        audioOutNotification = null,
        connection,
        micButton,
        micButtonEventRegistered,
        volumeButton,
        volumeRange,
        volumeRangeEventRegistered,
        volumeLabel;

    function setVolume (v) {
        var icon;

        log.debug("Client volume level set to " + v);

        settings.volume = v;
        settings.persist();
        connection.setVolume(v);

        if (v <= 0) {
            icon = icons["audio-off"];
        } else if (v <= 33) {
            icon = icons["audio-low"];
        } else if (v <= 66) {
            icon = icons["audio-medium"];
        } else {
            icon = icons["audio-on"];
        }

        volumeButton.find(".photon-icon").replaceWith(icon);
    }

    function toggleMicrophone () {
        settings.microphone = !settings.microphone;
        log.debug("[Audio][UI] Setting microphone to " + (settings.microphone ? "enabled" : "disabled"));
        setMicrophoneUI(settings.microphone);
    }

    function setMicrophoneUI (value) {
        connection.setMicrophone(value).then(function () {
            if (value) {
                micButton.find(".photon-icon").replaceWith(icons.unmuted);
            } else {
                micButton.find(".photon-icon").replaceWith(icons.muted);
            }
        }).catch(function (err) {
            var msg;

            msg = "Impossible to set audio input to " + value + ". Disabling microphone. Error (code: " + err.code + "): " + err.message;
            log.warn(msg);
            settings.microphone = false;
            audioNotificationManager.addNotification({
                level: notificationManager.level.WARNING,
                text: msg,
                actions: [{
                    dismissable: true
                }]
            });
        });
    }

    function audioInEnabled (status) {
        if (status.enabled) {
            settings.microphone = false; // Always start with mic off

            if (!micButtonEventRegistered) {
                micButton.on("click", function () {
                    toggleMicrophone();
                });
                micButtonEventRegistered = true;
            }

            micButton.show();

            if (audioInNotification !== null) {
                audioInNotification.remove();
                audioInNotification = null;
            }
        } else {
            setMicrophoneUI(false);
            micButton.hide();
            micButton.off("click");
            micButtonEventRegistered = false;

            if (audioOutNotification !== null) {
                // Display only audio-out notification if any

                if (audioInNotification !== null) {
                    audioInNotification.remove();
                    audioInNotification = null;
                }
            } else {
                audioInNotification = audioNotificationManager.addNotification({
                    level: notificationManager.level.WARNING,
                    text: "This browser or session does not support audio input",
                    actions: [{
                        dismissable: true
                    }]
                }, audioInNotification);
            }
        }
    }

    function audioOutEnabled (status) {
        if (status.enabled) {
            // FIXME: use bootstrap-slider to make it look good
            volumeRange = $("#dcv-volume-range").val(settings.volume);
            volumeLabel.val(settings.volume);

            if (!volumeRangeEventRegistered) {
                volumeRange.on("change", function () {
                    setVolume(volumeRange.val());
                }).on("input", function () {
                    volumeLabel.val(volumeRange.val());
                });
                volumeRangeEventRegistered = true;
            }

            volumeButton.show();

            if (audioOutNotification !== null) {
                audioOutNotification.remove();
                audioOutNotification = null;
            }
        } else {
            volumeButton.hide();
            setMicrophoneUI(false);
            micButton.hide();
            volumeRange.off("change");
            volumeRange.off("input");
            micButton.off("click");
            volumeRangeEventRegistered = false;

            // Display only audio-out notification
            audioOutNotification = audioNotificationManager.addNotification({
                level: notificationManager.level.WARNING,
                text: "Audio is not available",
                actions: [{
                    dismissable: true
                }]
            }, audioOutNotification);

            if (audioInNotification !== null) {
                audioInNotification.remove();
                audioInNotification = null;
            }
        }
    }

    return {
        setup: function (conn) {
            connection = conn;
            audioNotificationManager = notificationManager.getCategoryHandler("audio");
            featurehandler.addFeatureListener("audio-in", audioInEnabled);
            featurehandler.addFeatureListener("audio-out", audioOutEnabled);
            log = loglevel.getLogger("audioui");
            volumeButton = $("#dcv-volume-button");
            micButton = $("#dcv-mic-button");
            volumeLabel = $("#dcv-volume-label");
            micButtonEventRegistered = false;
            volumeRangeEventRegistered = false;
        }
    };
});

/* ex:set ts=4 et: */
