"use strict";

define([
    "jquery",
    "loglevel",
    "./utils",
    "./featurehandler",
    "bowser"
],
function ($, loglevel, utils, featurehandler, bowser) {
    var log,
        conn,
        clipboardUIButton,
        clipboardUICopyModal,
        clipboardUIPasteModal,
        clipboardUICopyMenuEntry,
        clipboardData,
        copyEnabled = false,
        pasteEnabled = false;

    function dataSizeWarning () {
        // we must destroy the previous popover to show the new one otherwise we'll see the old one
        clipboardUIButton.popover("destroy").popover({
            placement: "bottom",
            trigger: "manual",
            html: true,
            content: "WARNING: The clipboard data is too big to be transferred."
        });
        clipboardUIButton.popover("show");
        setTimeout(function () {
            clipboardUIButton.popover("hide");
        }, 2000);
    }

    function blinkIcon () {
        // make the button blink
        clipboardUIButton.fadeOut(150).fadeIn(150).fadeOut(150).fadeIn(150).fadeOut(150).fadeIn(150);
    }

    function newDataAvailableMessage () {
        clipboardUIButton.popover("destroy").popover({
            placement: "bottom",
            trigger: "manual",
            html: true,
            content: "Use the clipboard to copy data locally or remotely."
        });
        clipboardUIButton.popover("show");
        setTimeout(function () {
            clipboardUIButton.popover("hide");
        }, 2000);
    }

    function onMenuClickCopy (e) {
        if (conn.syncClipboards(e)) {
            log.debug("[UI][Clipboard] Copy done via menu click.");
            hideModalDialog(clipboardUICopyModal);
        } else {
            log.debug("[UI][Clipboard] Error on autocopy. Showing old copy modal.");
        }
    }

    function hideModalDialog (clipboardDialogModal) {
        // Add here effects on copy/paste success

        $(".dcv-clipboard-success", clipboardDialogModal).show();
        $("textarea", clipboardDialogModal).val("");
        setTimeout(function () {
            clipboardDialogModal.modal("hide");
            $(".dcv-clipboard-success", clipboardDialogModal).hide();
        }, 1250);
    }

    function onModalHide (e) {
        conn.captureClipboardEvents(false);
        log.debug("[UI] Hiding clipboard modal");
    }

    function onModalCopyShow (e) {
        var remoteTextData;

        if (clipboardData) {
            remoteTextData = clipboardData["text/plain"];

            if (remoteTextData) {
                // FIXME: This should replaced with a better fitting function to avoid
                // chars outside the text-area (ie the case with a lot of end lines)
                // Moreover, the following avoids to print the end string char needed at C level
                $("textarea", clipboardUICopyModal).val((remoteTextData.substring(0, 100)).replace(/\0/g, ""));
            }
        }

        $("textarea", clipboardUICopyModal).prop("disabled", true);
        log.debug("[UI] Showing clipboard modal");
    }

    function onModalCopyShowDone (e) {
        onModalCopyShow(e);
        setTimeout(function () {
            $("#dcv-clipboard-copy").hide();
        }, 1000);
    }

    function onModalPasteShow (e) {
        // No specific action for now in the paste modal
        log.debug("[UI] Showing clipboard modal");
    }

    function onModalShown () {
        conn.captureClipboardEvents(true);
        log.debug("[UI] Clipboard modal shown");
    }

    function showInterface () {
        if (copyEnabled || pasteEnabled) {
            clipboardUIButton.show();
        }
    }

    function addCopyListeners (autoCopyAvailable) {
        if (autoCopyAvailable) {
            clipboardUICopyMenuEntry.on("click", onMenuClickCopy);
            clipboardUICopyModal.on("hide.bs.modal", onModalHide);
            clipboardUICopyModal.on("show.bs.modal", onModalCopyShowDone);
            clipboardUICopyModal.on("shown.bs.modal", onModalShown);
        } else {
            clipboardUICopyModal.on("hide.bs.modal", onModalHide);
            clipboardUICopyModal.on("show.bs.modal", onModalCopyShow);
            clipboardUICopyModal.on("shown.bs.modal", onModalShown);
        }
    }

    function removeCopyListeners () {
        clipboardUICopyMenuEntry.off("click", onMenuClickCopy);
        clipboardUICopyModal.off("hide.bs.modal", onModalHide);
        clipboardUICopyModal.off("show.bs.modal", onModalCopyShowDone);
        clipboardUICopyModal.off("shown.bs.modal", onModalShown);
        clipboardUICopyModal.off("show.bs.modal", onModalCopyShow);
    }

    function addPasteListeners (autoPasteAvailable) {
        clipboardUIPasteModal.on("hide.bs.modal", onModalHide);
        clipboardUIPasteModal.on("show.bs.modal", onModalPasteShow);
        clipboardUIPasteModal.on("shown.bs.modal", onModalShown);
    }

    function removePasteListeners (autoPasteAvailable) {
        clipboardUIPasteModal.off("hide.bs.modal", onModalHide);
        clipboardUIPasteModal.off("show.bs.modal", onModalPasteShow);
        clipboardUIPasteModal.off("shown.bs.modal", onModalShown);
    }

    function copyFeatureEnabled (status) {
        if (!status) {
            return;
        }

        if (copyEnabled !== status.enabled) {
            copyEnabled = status.enabled;
            showInterface();
            removeCopyListeners();
            if (copyEnabled) {
                $("#dcv-copy").show();
                addCopyListeners(status.autoCopy);
            } else {
                $("#dcv-copy").hide();
            }
        }
    }

    function pasteFeatureEnabled (status) {
        if (!status) {
            return;
        }

        if (pasteEnabled !== status.enabled) {
            pasteEnabled = status.enabled;
            showInterface();
            removePasteListeners();
            if (pasteEnabled) {
                $("#dcv-paste").show();
                addPasteListeners(status.autoPaste);
            } else {
                $("#dcv-paste").hide();
            }
        }
    }

    return {
        setup: function (connection, button, copyModal, pasteModal, copyEntry) {
            log = loglevel.getLogger("clipboardUI");

            conn = connection;
            clipboardUIButton = button; // we assume that is a jquery object
            clipboardUICopyModal = copyModal; // we assume that is a jquery object
            clipboardUIPasteModal = pasteModal; // we assume that is a jquery object
            clipboardUICopyMenuEntry = copyEntry; // we assume that is a jquery object
            $("#dcv-copy").hide();
            $("#dcv-paste").hide();
            clipboardUIButton.hide();
            featurehandler.addFeatureListener("clipboard-copy", copyFeatureEnabled);
            featurehandler.addFeatureListener("clipboard-paste", pasteFeatureEnabled);
        },

        clipboardEvent: function (eventToHandle) {
            log.debug("[UI][Clipboard] Received clipboard event " + eventToHandle.name);
            switch (eventToHandle.name) {
            case "established":
                // The following value is not used in the UI at the moment
                // maxDataSizeAllowed = eventToHandle.maxDataSize;
                showInterface();
                break;
            case "copy":
                hideModalDialog(clipboardUICopyModal);
                break;
            case "paste":
                hideModalDialog(clipboardUIPasteModal);
                break;
            case "dataSizeAlert":
                dataSizeWarning();
                break;
            case "autoCopyDone":
                blinkIcon();
                break;
            case "autoPasteDone":
                blinkIcon();
                break;
            case "remoteError":
                if (eventToHandle.error === "dataSizeAlert") {
                    dataSizeWarning();
                } // The other errors are not handle in the UI for now
                break;
            case "newDataAvailable":
                if (!eventToHandle.autoCopy) {
                    newDataAvailableMessage();
                }
                clipboardData = eventToHandle.clipboardData;
                break;
            default:
                log.debug("[UI][Clipboard] Unkown clipboard event.");
            }
        }
    };
});

/* ex:set ts=4 et: */
