"use strict";

define([
    "jquery",
    "loglevel",
    "bytebuffer",
    "saslfactory",
    "./proto/auth",
    "./icons",
    "./settings",
    "./utils",
    "./postsession",
    "text!templates/login.html",
    "text!images/dcv.svg",
    "bootstrap"
],
function ($, loglevel, ByteBuffer, saslfactory, authproto, icons, settings, utils, postsession, loginTemplate, logo) {

    var log,
        onSuccess,
        websocket,
        serverMechanisms,
        mechanism,
        credentials,
        loginDiv,
        usernameInput,
        passwordInput,
        session,
        shiftPressed = false,
        capsOn = false,
        permanentErrorShown = false;

    function sendMessage (json) {
        var msg = new authproto.dcv.auth.ClientMessage(json),
            msgBuffer = authproto.dcv.auth.ClientMessage.encode(msg).finish(),
            msgSize = msgBuffer.byteLength,
            paddedMsgSize = (msgSize + 7) & (~7),
            frame = new Uint8Array(8 + paddedMsgSize),
            dv = new DataView(frame.buffer);

        // Use the same framing of the DCV protocol, but
        // without binary payload
        dv.setUint32(0, msgSize, true);
        dv.setUint32(4, 0, true);
        frame.set(new Uint8Array(msgBuffer), 8);

        try {
            websocket.send(frame.buffer);
        } catch (err) {
            log.warn("Failed to send: " + err.message);
        }
    }

    function onOpen () {
        sendMessage({
            authenticationRequest: {}
        });
    }

    function onClose (e) {
        log.debug("auth websocket closed. Code: " + e.code);
        showError("Connection closed (timeout)", true);
    }

    function onMessage (e) {
        var dv = new DataView(e.data),
            msgSize = dv.getUint32(0, true),
            binSize = dv.getUint32(4, true),
            msg;

        if (binSize !== 0) {
            log.error("Invalid message (bin size: " + binSize + ")");
            websocket.close();
        }

        msg = authproto.dcv.auth.ServerMessage.decode(new Uint8Array(e.data.slice(8, 8 + msgSize)));

        switch (msg.msg) {
        case "authenticationResult":
            handleAuthenticationResult(msg.authenticationResult);
            break;
        case "authenticationInit":
            handleAuthenticationInit(msg.authenticationInit);
            break;
        case "saslChallenge":
            handleSaslChallenge(msg.saslChallenge);
            break;
        default:
            log.error("Unexpected message:" + msg.msg);
            websocket.close();
        }
    }

    function onError () {
        showError("Failed to communicate with server", true);
    }

    function handleAuthenticationResult (msg) {
        var i = null,
            message;

        if (msg.resultValue === authproto.dcv.auth.AuthenticationResult.Result.OK) {
            if (msg.sessionTokens.length > 0) {
                // FIXME: show selection UI for the available sessions, for now we take the first
                if (utils.isFunction(onSuccess)) {
                    if (session === "") {
                        i = 0;
                    } else {
                        msg.sessionTokens.forEach(function (token, index) {
                            if (token.sessionId === session) {
                                i = index;
                            }
                        });
                    }

                    loginDiv.remove();
                    if (i === null) {
                        message = "The session " + session + " is not available";
                        log.error(message);
                        postsession.showError({message: message}, false);
                    } else {
                        onSuccess(msg.sessionTokens[i].sessionId, msg.sessionTokens[i].authenticationToken);
                    }
                }
            } else {
                log.error("No sessions available");
                showError("No sessions available", true);
            }
            websocket.close();
        } else if (msg.retriesLeft > 0) {
            log.debug("Wrong username or password");
            showError("Wrong username or password", false);
        } else {
            log.error("Authentication failed");
            showError("Authentication failed", true);
        }
    }

    function handleAuthenticationInit (msg) {
        var urlParams, user, pass;

        urlParams = utils.getUrlParams();
        user = urlParams.username || "";
        pass = urlParams.password || "";
        serverMechanisms = msg.mechanisms;

        utils.clearUrlParam("username");
        utils.clearUrlParam("password");

        switch (msg.authenticationMode) {
        case "dcv":
        case "system":
            if ((user !== "") && (pass !== "")) {
                credentials = {username: user, password: pass};
                saslStart();
                return;
            }
            usernameInput = addUserNameInput(user);
            passwordInput = addPasswordInput(pass);
            break;
        case "otp":
            if (pass !== "") {
                credentials = {username: "dcvuser", password: pass};
                saslStart();
                return;
            }
            passwordInput = addPasswordInput(pass);
            break;
        case "none":
            credentials = {username: "dcvuser", password: ""};
            saslStart();
            return;
        default:
            showError("Authentication failed", true);
            log.error("Unknown authentication mode " + msg.mode);
            websocket.close();
            return;
        }

        $("form", loginDiv).submit(function (event) {
            onSubmit(event);
            event.preventDefault();
            return false;
        });

        loginDiv.fadeIn();
    }

    function handleSaslChallenge (msg) {
        var challenge;

        if (msg.authData.length > 0) {
            challenge = String.fromCharCode.apply(null, msg.authData);
            log.debug("SASL challenge: " + challenge);
            if (!mechanism.challenge(challenge)) {
                showError("Authentication failed", true);
                log.info(mechanism.error);
                websocket.close();
                return;
            }
        }

        if (mechanism.completed) {
            log.error("Unexpected SASL protocol state");
            websocket.close();
            return;
        }

        saslStep();
    }

    function getAuthData () {
        var response = mechanism.response(credentials),
            authData = new ByteBuffer();

        authData.writeUTF8String(response);
        authData.flip();

        return new Uint8Array(authData.toBuffer(true));
    }

    function saslStart () {
        log.debug("SASL start");

        mechanism = saslfactory.create(serverMechanisms);
        if (!mechanism) {
            showError("Authentication failed", true);
            log.error("Impossible to find a suitable SASL authentication mechanism in: " + serverMechanisms);
            websocket.close();
            return;
        }

        log.debug("Using SASL mechanism: " + mechanism.name);

        if (!mechanism.clientFirst) {
            log.error("Unable to set data for SASL mechanism " + mechanism.name);
            websocket.close();
            return;
        }

        sendMessage({
            saslStart: {
                mechanism: mechanism.name,
                authData: getAuthData()
            }
        });
    }

    function saslStep () {
        log.debug("SASL step");

        sendMessage({
            saslStep: {
                authData: getAuthData()
            }
        });
    }

    function checkCapsLock (entry) {
        entry.bind("keydown", function (e) {
            if (e.key === "Shift") {
                shiftPressed = true;
            }
        }).bind("keyup", function (e) {
            if (e.key === "Shift") {
                shiftPressed = false;
            } else if (e.key === "CapsLock") {
                // we cannot reliably detect that pressing caps lock
                // turns it on, we can only detect that if it was on
                // because pressing "a" resuts in "A" then pressing
                // caps lock turns it off
                if (capsOn) {
                    capsOn = false;
                    $(".dcv-check-caps").popover("hide");
                }
            }
        }).bind("keypress", function (e) {
            var c, lowercase, newCapsOn;

            // assume normal keys only have one codepoint
            if (e.key && !e.key.codePointAt(1)) {
                c = e.key.charAt(0);
                if (c) {
                    lowercase = c.toLowerCase();
                    newCapsOn = !shiftPressed && (c !== lowercase);

                    if (capsOn !== newCapsOn) {
                        $(".dcv-check-caps").popover("toggle");
                        capsOn = newCapsOn;
                    }
                }
            }
        });
    }

    function addUserNameInput (value) {
        var input;

        input = $("<input name=\"username\" type=\"text\" value=\"" + value + "\" class=\"form-control auth-token\" placeholder=\"Username\">").
            appendTo($("fieldset", loginDiv));

        checkCapsLock(input);

        return input;
    }

    function addPasswordInput (value) {
        var input;

        input = $("<input name=\"password\" type=\"password\" value=\"" + value + "\" class=\"form-control auth-token dcv-check-caps\" placeholder=\"Password\">").
            appendTo($("fieldset", loginDiv));

        input.popover({
            placement: "right",
            trigger: "manual",
            html: true,
            content: "Caps Lock is on"
        });

        checkCapsLock(input);

        return input;
    }

    function encodeUtf8 (s) {
        return unescape(encodeURIComponent(s));
    }

    /* FIXME: prevent multiple submissions */
    function onSubmit (event) {
        var username, password;

        password = encodeUtf8(passwordInput.val());
        if (usernameInput) {
            username = encodeUtf8(usernameInput.val());
        } else {
            username = "dcvuser";
        }

        if (username.length === 0) {
            usernameInput.focus();
            return;
        }

        if (password.length === 0) {
            passwordInput.focus();
            return;
        }

        credentials = {username: username, password: password};
        saslStart();
    }

    function showError (errorMessage, permanent) {
        if (permanent) {
            if (permanentErrorShown) {
                return;
            }

            permanentErrorShown = true;
        }

        $(".dcv-login-message", loginDiv).addClass("error");
        $(".dcv-login-message .photon-icon", loginDiv).replaceWith(icons["error-notification"]);
        $(".dcv-login-message span", loginDiv).text(errorMessage);
        $("input.auth-token", loginDiv).val("");
        if (permanent) {
            $("form", loginDiv).css("visibility", "hidden");
        }
    }

    return {
        show: function (config) {
            log = loglevel.getLogger("loginform");
            onSuccess = config.onSuccess;
            session = config.requestedSession;

            loginDiv = $(loginTemplate).appendTo(document.body).hide();

            $(".dcv-logo", loginDiv).append(logo);

            websocket = new WebSocket(config.url, "dcv");
            websocket.binaryType = "arraybuffer";
            websocket.onopen = onOpen;
            websocket.onclose = onClose;
            websocket.onmessage = onMessage;
            websocket.onerror = onError;
        }
    };
});

/* ex:set ts=4 et: */
