"use strict";

define([
    "loglevel",
    "dcv",
    "./audio",
    "./clipboard",
    "./printer",
    "./loginform",
    "./postsession",
    "./featurehandler",
    "./filestorage",
    "./viewer",
    "./settings",
    "./licensehandler",
    "./utils",
    "bowser"
],
function (loglevel, dcv, audio, clipboard, printer, loginform, postsession, featurehandler, filestorage, viewer, settings, licensehandler, utils, bowser) {
    var log;

    document.title = location.hostname + " - DCV";

    settings.load();
    initLogLevel();

    log.info("Using DCV JS " + dcv.version.versionStr);

    function initLogLevel () {
        var level,
            dcvLogLevel,
            domainsLogLevel,
            defaultLogLevel = "info";

        if (settings.logLevel.indexOf("@") !== -1) {
            // we have a dictionary for log levels
            dcvLogLevel = {};
            domainsLogLevel = settings.logLevel.split(",");
            domainsLogLevel.forEach(function (e) {
                if (e.indexOf("@") !== -1) {
                    level = e.split("@");
                    dcvLogLevel[level[0]] = level[1];
                } else {
                    dcvLogLevel.default = e;
                    defaultLogLevel = e;
                }
            });
            switch (defaultLogLevel.toLowerCase()) {
            case "trace":
                level = loglevel.levels.TRACE;
                break;
            case "debug":
                level = loglevel.levels.DEBUG;
                break;
            case "warn":
                level = loglevel.levels.WARN;
                break;
            case "error":
                level = loglevel.levels.ERROR;
                break;
            case "silent":
                level = loglevel.levels.SILENT;
                break;
            case "info":
            default:
                level = loglevel.levels.INFO;
                break;
            }
        } else {
            switch (settings.logLevel.toLowerCase()) {
            case "trace":
                level = loglevel.levels.TRACE;
                dcvLogLevel = dcv.LogLevel.TRACE;
                break;
            case "debug":
                level = loglevel.levels.DEBUG;
                dcvLogLevel = dcv.LogLevel.DEBUG;
                break;
            case "warn":
                level = loglevel.levels.WARN;
                dcvLogLevel = dcv.LogLevel.WARN;
                break;
            case "error":
                level = loglevel.levels.ERROR;
                dcvLogLevel = dcv.LogLevel.ERROR;
                break;
            case "silent":
                level = loglevel.levels.SILENT;
                dcvLogLevel = dcv.LogLevel.SILENT;
                break;
            case "info":
            default:
                level = loglevel.levels.INFO;
                dcvLogLevel = dcv.LogLevel.INFO;
                break;
            }
        }

        loglevel.setLevel(level);
        dcv.setLogLevel(dcvLogLevel);

        // sample log handler registration
        /*
        dcv.setLogHandler(function (m) {
            console.log("[CUSTOM-LOG-HANDLER] [" + m.levelName + " (" + m.level + ")] [" + m.domain + "] " + m.message);
        });
        */

        log = loglevel.getLogger("main");
    }

    function getEndpoint (route) {
        var scheme = settings.wss ? "wss" : "ws",
            host = window.location.hostname,
            port = window.location.port,
            path = window.location.pathname.replace(/\/[^/]*$/, "");

        return scheme + "://" + host + (port ? ":" + port : "") + path + route;
    }

    function connect (sessionId, authToken) {
        viewer.showDisplayFrame();
        log.info("Calling client.connect");
        dcv.connect({
            url: getEndpoint("/ws"),
            sessionId: sessionId,
            authToken: authToken,
            enabledChannels: settings.enabledChannels,
            displayCodecs: settings.displayCodecs,
            divId: "dcv-display",
            showDebugLayer: settings.debugLayer,
            baseUrl: "js/lib/dcv",
            resourceBaseUrl: settings.resourceBaseUrl,
            volumeLevel: settings.volume,
            audioBitrate: settings.audioBitrate,
            fingerRadius: settings.fingerRadius,
            clipboardAutoSync: true,
            observers: {
                displayLayout: function (serverdWidth, serverHeight) {
                    viewer.setDisplayLayout(serverdWidth, serverHeight);
                },
                firstFrame: function (resizeEnabled, relativeMouseModeEnabled) {
                    viewer.showDisplay(resizeEnabled, relativeMouseModeEnabled);
                },
                disconnect: function (reason) {
                    handleDisconnect(reason);
                },
                licenseNotification: function (notification) {
                    licensehandler.handleNotification(notification);
                },
                idleWarningNotification: function (disconnectionDateTime) {
                    viewer.handleIdleTimeoutWarning(disconnectionDateTime);
                },
                dataChannel: function (info) {
                    viewer.customChannel(info);
                },
                filePrinted: function (printResource) {
                    printer.handleResource(printResource);
                },
                fileDownload: function (fileResource) {
                    filestorage.downloadResource(fileResource);
                },
                qualityIndicatorState: function (indicators) {
                    log.info("Quality Indicators: " + JSON.stringify(indicators));
                    viewer.qualityIndicatorChange(indicators);
                },
                filestorageEnabled: function (enabled) {
                    filestorage.enabled(enabled);
                },
                clipboardEvent: function (eventToHandle) {
                    clipboard.clipboardEvent(eventToHandle);
                },
                featuresUpdate: function (list) {
                    featurehandler.featuresUpdate(list);
                }
            }
        }).then(function (connection) {
            viewer.setConnection(connection);
        }).catch(function (error) {
            log.error("client.connect was rejected (" + error + ")");
            viewer.remove();
            postsession.showError(error, settings.promptReconnect);
        });
    }

    function showLoginForm () {
        loginform.show({
            url: getEndpoint("/auth"),
            requestedSession: settings.sessionId,
            onSuccess: connect
        });
    }

    function handleDisconnect (reason) {
        viewer.remove();

        /* At the moment we do not expose any specific message on the error
         * except for idle-disconnect. The message will be shown after
         * 'The connection has been closed', the message should be 30 chars
         * log since we have only 1 line for that in the UI.
         */
        switch (reason.code) {
        case dcv.ClosingReasonError.IDLE_TIMEOUT_EXPIRED:
            postsession.showReason({code: reason.code, message: "The session has been marked as idle"}, false, settings.promptReconnect);
            break;
        default:
            postsession.showReason(null, false, settings.promptReconnect);
        }
    }

    // FIXME: these are not real settings, but this a quick
    // way to bypass internal auth and use an external auth token
    if (settings.authToken) {
        if (!settings.promptReconnect) {
            utils.clearUrlParam("authToken");
        }

        if (settings.sessionId) {
            log.info("Logging in using authenticatin token (" + settings.authToken + ") and session id (" + settings.sessionId + ")");
            connect(settings.sessionId, settings.authToken);
        } else {
            log.info("Logging in using authenticatin token (" + settings.authToken + ") and default session id (console)");
            connect("console", settings.authToken);
        }
    } else {
        log.info("Showing log in form");
        showLoginForm();
    }
});

/* ex:set ts=4 et: */
