"use strict";

define([
    "jquery",
    "loglevel",
    "dcv",
    "./audio",
    "./clipboard",
    "./featurehandler",
    "./filestorage",
    "./fullscreen",
    "./icons",
    "./printer",
    "./settings",
    "./licensehandler",
    "./utils",
    "./console",
    "./keyboardshortcuts",
    "./notificationmanager",
    "./softkeyboard",
    "text!templates/viewer.html",
    "text!images/dcv.svg",
    "bowser",
    "bootstrap" // does not return an object, must be after the other modules
],
function ($, loglevel, dcv, audio, clipboard, featurehandler, filestorage, fullscreen, icons, printer, settings, licensehandler, utils, clientConsole, keyboardshortcuts, notificationManager, softkeyboard, viewerTemplate, logo, bowser) {
    var log,
        container = $(viewerTemplate),
        notificationsButton = $("#dcv-notifications-button", container),
        notificationsContainer = $("#dcv-notifications-container", container),
        qiNotification = null,
        idleNotification = null,
        displayFrame = $("#dcv-display-frame", container),
        displayDiv = $("#dcv-display", container),
        connection,
        resizeTimer,
        builtinResolutions = [
            {width: 800, height: 600},
            {width: 1024, height: 768},
            {width: 1440, height: 900},
            {width: 1920, height: 1080}
        ],
        requestedResolution = {width: 1, height: 1};

    function toggleFullscreen () {
        if (fullscreen.element) {
            fullscreen.exit();
        } else {
            fullscreen.request(container.get(0));
        }
    }

    function resizeDisplay () {
        var height = window.innerHeight,
            margin;

        if (!fullscreen.element) {
            height -= $(".dcv-toolbar-container", container).outerHeight(true);
        }
        displayFrame.height(height + "px");
        margin = Math.max(0, (height - displayDiv.height()) / 2);
        displayDiv.css("margin-top", margin + "px");

        // force a redraw of the div, this fix a bug in chrome which cause the use of scrollbars even if not needed
        displayFrame.hide().show(0);
    }

    function requestDisplayLayoutChange () {
        resizeDisplay();

        if (settings.dynamicResolution) {
            if (resizeTimer) {
                clearTimeout(resizeTimer);
            }
            if (displayFrame.width() !== displayDiv.width() || displayFrame.height() !== displayDiv.height()) {
                resizeTimer = setTimeout(function (w, h) {
                    requestResolution(w, h).catch(function (err) {
                        log.warn("Failed to request resolution " + w + "x" + h + ": " + err.message);
                    });
                    resizeTimer = null;
                }, 1000, displayFrame.width(), displayFrame.height());
            }
        }
    }

    function setDisplayLayout (serverdWidth, serverHeight) {
        var scaleRatio = 1;

        resizeDisplay();

        if (settings.scaleToFit) {
            scaleRatio = Math.floor(Math.min(1, requestedResolution.width / serverdWidth, requestedResolution.height / serverHeight) * 100) / 100;
            if (!isNaN(scaleRatio) && scaleRatio > 0 && scaleRatio <= 1) {
                log.info("[UI] Requesting a display scaling factor of " + scaleRatio);
                if (setCanvasScale(scaleRatio)) {
                    connection.setDisplayScale(scaleRatio).then(function () {
                        resizeDisplay();
                    }).catch(function (err) {
                        log.warn("Failed to request scaling factor of " + scaleRatio + ": " + err.message);
                        setCanvasScale(1);
                    });
                } else {
                    log.info("[UI] CSS scaling or zoom unsupported!");
                }
            }
        }
    }

    function setCanvasScale (z) {
        var scale = "scale(" + z + ")",
            supportsCSS = (window.CSS && window.CSS.supports) || window.supportsCSS;

        if (supportsCSS && supportsCSS("( transform: scale(1) )")) {
            displayDiv.css("transform", scale);
            displayDiv.css("transform-origin", "0 0");
            displayFrame.css("overflow", "hidden"); // Since we are scaling we should not show scrollbars
            return true;
        }

        if (supportsCSS && supportsCSS("( zoom: 0.8 )")) {
            displayDiv.css("zoom", z);
            displayFrame.css("overflow", "hidden"); // Since we are scaling we should not show scrollbars
            return true;
        }

        displayFrame.css("overflow", "auto"); // No scale, thus we can have scrollbars
        return false;
    }

    function setResolutionIcon (elem, dynamicResolution) {
        $("#dcv-resolution-list", container).children("li").find("svg").attr("class", "photon-icon photon-icon-checked place-holder-icon");
        elem.find("svg").attr("class", "photon-icon photon-icon-checked");

        settings.dynamicResolution = dynamicResolution;
        settings.persist();
    }

    function showSpinner () {
        /* Append the spinner to the body so that the animation is not affected by the display resize */
        $(document.body).append("<div class='dcv-spinner'/>");
    }

    function removeSpinner () {
        $(".dcv-spinner").remove();
    }

    function renderServerInfo (info, baseText) {
        var serverInfo = "";

        Object.keys(info).forEach(function (el) {
            if (typeof info[el] === "object") {
                serverInfo += renderServerInfo(info[el], el);
            } else if (!baseText) {
                serverInfo = serverInfo + el + ": " + info[el] + "\n";
            } else {
                serverInfo = serverInfo + baseText + "-" + el + ": " + info[el] + "\n";
            }
        });
        return serverInfo;
    }

    function setAboutDialog () {
        $("#dcv-about-client").find("textarea").html(
            "OS: " + bowser.osname + " " + bowser.osversion + "\n" +
            "Browser version: " + bowser.name + " " + bowser.version + "\n" +
            "Client version: " + dcv.version.versionStr);
        $("#dcv-about-server").find("textarea").html(
            "SERVER\n" + renderServerInfo(connection.getServerInfo()));
        $("#dcv-about-third").find("textarea").load("licenses.txt");
        $("#dcv-about-footer").html(
            "<br/><a href=\"EULA.txt\" target=\"_blank\">End User License Agreement</a>" +
            "<br/>&copy 2018, Nice Srl. All rights reserved<br/>");
        $(".dcv-logo").append(logo);

        $(".tabs .tab-links a").on("click", function (e) {
            /* eslint-disable no-invalid-this*/
            var currentAttrValue = "#" + $(this).attr("data-target");

            /* Show/Hide tabs */
            $(currentAttrValue).show().siblings().hide();

            /* Change/remove current tab to active */
            $(this).parent("li").addClass("active").siblings().removeClass("active");

            /* eslint-enable no-invalid-this*/
            e.preventDefault();
        });
    }

    function requestResolution (width, height) {
        requestedResolution.width = width;
        requestedResolution.height = height;
        return connection.requestResolution(width, height);
    }

    // FIXME: I think we should use the *system* notification
    // instead of the dcv one since we want to get the attention
    // from the user even when he is not looking at the current tab
    function handleIdleTimeoutWarning (disconnectionTimeDate) {
        var idleNotificationHandler,
            messageText;

        if (disconnectionTimeDate) {
            messageText = "DCV connection is idle and is going to be disconnected at " + disconnectionTimeDate.toString();
            log.info(messageText);
            idleNotificationHandler = notificationManager.getCategoryHandler("idle warning");

            idleNotification = idleNotificationHandler.addNotification({
                level: notificationManager.level.WARNING,
                text: messageText,
                actions: [{
                    dismissable: true
                }]
            }, idleNotification);
        } else if (idleNotification) {
            idleNotification.remove();
            idleNotification = null;
        }
    }

    return {
        showDisplayFrame: function () {
            log = loglevel.getLogger("viewer");
            container = container.appendTo(document.body);
            showSpinner();

            /* Initilize Bootstrap widgets here */
        },

        setConnection: function (conn) {
            connection = conn;
            featurehandler.setup(connection);
            notificationManager.setup(notificationsButton, notificationsContainer, container); // Must be the first UI element enabled
            licensehandler.setup();
            clientConsole.setup($("#dcv-console", container), connection, settings.minQuality);
            clipboard.setup(connection, $("#dcv-clipboard-button", container), $("#dcv-clipboard-copy", container), $("#dcv-clipboard-paste", container), $("#dcv-copy", container));
            filestorage.setup(connection);
            audio.setup(connection);
            softkeyboard.setup(connection, $("#dcv-softkeyboard-button", container));
            printer.setup();
            keyboardshortcuts.setup($("#dcv-shortcuts-editor", container), connection);
        },

        showDisplay: function (resizeEnabled, relativeMouseModeEnabled) {
            var fullscreenButton = $("#dcv-fullscreen-button", container),
                matchClientResolution = $("#dcv-match-local-resolution", container),
                keepServerResolution = $("#dcv-keep-server-resolution", container),
                relativeMouseButton = $("#dcv-relative-mouse-position");

            removeSpinner();

            icons.setupPhotonIcons();

            $("#dcv-host-label", container).text(connection.getServerInfo().hostname);

            setAboutDialog();

            // initially the toolbar is invisible (visibility: hidden) but not collapsed,
            // since we do not want to cause a resize of the display area at the first frame
            $(".dcv-toolbar-container").removeClass("invisible");

            displayDiv.css("margin", "0 auto").css("position", "relative");
            window.addEventListener("resize", function (e) {
                requestDisplayLayoutChange();
            });

            featurehandler.addFeatureListener("keyboard-sas", function (status) {
                if (status.enabled) {
                    $("#dcv-sas", container).show(0);
                } else {
                    $("#dcv-sas", container).hide();
                }
            });

            $("#dcv-sas", container).click(function () {
                connection.sendKeyboardShortcut([
                    {key: "Control", location: KeyboardEvent.DOM_KEY_LOCATION_LEFT},
                    {key: "Alt", location: KeyboardEvent.DOM_KEY_LOCATION_LEFT},
                    {key: "Delete", location: KeyboardEvent.DOM_KEY_LOCATION_STANDARD}
                ]);
            });

            $("#dcv-disconnect", container).click(function () {
                connection.disconnect();
            });

            window.addEventListener("beforeunload", function () {
                connection.disconnect();
            });

            if (resizeEnabled) {
                if (settings.dynamicResolution) {
                    setResolutionIcon(matchClientResolution, true);
                } else {
                    setResolutionIcon(keepServerResolution, false);
                }

                matchClientResolution.click(function () {
                    setResolutionIcon(matchClientResolution, true);
                    requestDisplayLayoutChange();
                });

                keepServerResolution.click(function () {
                    setResolutionIcon(keepServerResolution, false);
                });

                builtinResolutions.filter(function (res) {
                    var resToKeep = true;

                    if (settings.minDisplayWidth) {
                        resToKeep = resToKeep && res.width >= settings.minDisplayWidth;
                    }
                    if (resToKeep && settings.maxDisplayWidth) {
                        resToKeep = resToKeep && res.width <= settings.maxDisplayWidth;
                    }
                    if (resToKeep && settings.minDisplayHeight) {
                        resToKeep = resToKeep && res.height >= settings.minDisplayHeight;
                    }
                    if (resToKeep && settings.maxDisplayWidth) {
                        resToKeep = resToKeep && res.height <= settings.maxDisplayHeight;
                    }

                    return resToKeep;
                }).forEach(function (res) {
                    var resolutionListPlaceholder = $("#dcv-resolutions-placeholder", container),
                        resolutionList = resolutionListPlaceholder.parent();

                    $("<a href='#'>" + icons.checked.replace("photon-icon-checked", "photon-icon-checked place-holder-icon") + res.width + "x" + res.height + "</a>").click(function () {
                        if (connection) {
                            requestResolution(res.width, res.height);

                            /* eslint-disable no-invalid-this */
                            setResolutionIcon($(this), false);

                            /* eslint-enable no-invalid-this */
                        } else {
                            log.warn("Attempted request resolution without an active connection.");
                        }
                    }).appendTo(resolutionList).wrap("<li></li>");

                });
            } else {
                /* User not allowed to change resolution, hide the button */
                $("#dcv-resolution-button", container).hide();
            }

            if (relativeMouseModeEnabled) {
                relativeMouseButton.click(function () {
                    connection.enterRelativeMouseMode();
                });
            } else {
                relativeMouseButton.hide();
            }

            if (fullscreen.enabled) {
                fullscreenButton.click(function () {
                    toggleFullscreen();
                });

                fullscreen.addEventListener(function (event) {
                    container.toggleClass("fullscreen");

                    if (fullscreen.element) {
                        $("#dcv-toolbar-collapse").collapse("hide");
                        $(".photon-icon", fullscreenButton).replaceWith(icons["leave-full-screen"]);
                    } else {
                        $("#dcv-toolbar-collapse").collapse("show");
                        $(".photon-icon", fullscreenButton).replaceWith(icons["full-screen"]);
                    }
                });
            } else {
                fullscreenButton.hide();
            }

            // Since this is just for testing, send the settings message
            // only to turn QU off and do not send anything by default
            if (!settings.enableQU) {
                connection.enableDisplayQualityUpdates(settings.enableQU);
            }

            connection.setDisplayQuality(settings.minQuality, settings.maxQuality);
            connection.setMaxDisplayResolution(settings.maxDisplayWidth, settings.maxDisplayHeight);
            connection.setMinDisplayResolution(settings.minDisplayWidth, settings.minDisplayHeight);

            requestDisplayLayoutChange();
        },

        remove: function () {
            removeSpinner();
            container.remove();
        },

        resizeDisplay: resizeDisplay,

        setDisplayLayout: setDisplayLayout,

        handleIdleTimeoutWarning: handleIdleTimeoutWarning,

        // This is just for demonstration purposes: if a custom
        // channel is enabled, we echo back every message we receive
        customChannel: function (channelInfo) {
            var echoFunc;

            if (settings.customChannels.indexOf(channelInfo.name) === -1) {
                log.info("[UI] Custom channel '" + channelInfo.name + "' not supported");
                return;
            }

            connection.openChannel(channelInfo.name, channelInfo.authToken, {
                onMessageReceived: function (binaryPayload) {
                    log.info("[UI] Custom channel: received message " + new TextDecoder("utf-8").decode(binaryPayload));
                    echoFunc(binaryPayload);
                },
                onClose: function () {
                    log.info("[UI] Echo: closed");
                }
            }).then(function (channel) {
                log.info("[UI] Opened custom channel");
                echoFunc = function (msg) {
                    channel.sendMessage(msg).then(function () {
                        log.info("[UI] Success sending message.");
                    }).catch(function (error) {
                        log.error("Failed to send message: " + JSON.stringify(error));
                    });
                };
            }).catch(function (error) {
                log.error("Failed to open custom channel: " + error.message);
            });
        },

        qualityIndicatorChange: function (indicators) {
            var qiNotificationHandler,
                messageText = "",
                alarmState = 0;

            qiNotificationHandler = notificationManager.getCategoryHandler("quality indicator");
            indicators.indicators.forEach(function (indicator) {
                messageText += "<strong>" + indicator.name + "</strong> status: " + indicator.status + "<br />";
                if (indicator.changed) {
                    log.info("[UI] Indicator " + indicator.name + " has changed to status " + indicator.status);
                }
                if (indicator.status > alarmState) {
                    alarmState = indicator.status;
                }
            });

            log.info("[UI] Current Quality Indicator overall alarm state: " + alarmState);

            switch (alarmState) {
            case 1:
                qiNotification = qiNotificationHandler.addNotification({
                    level: notificationManager.level.WARNING,
                    text: messageText,
                    actions: [{
                        dismissable: true
                    }]
                }, qiNotification);
                break;
            case 2:
                qiNotification = qiNotificationHandler.addNotification({
                    level: notificationManager.level.ERROR,
                    text: messageText,
                    actions: [{
                        dismissable: true
                    }]
                }, qiNotification);
                break;
            case 0:
            default:
                if (qiNotification) {
                    qiNotification.remove();
                    qiNotification = null;
                }
                break;
            }
        }
    };
});

/* ex:set ts=4 et: */
