/* eslint-disable new-cap */

"use strict";

define(["./bitops"], function (bitops) {

    var RESP = {},
        CLIENT_KEY = "Client Key",
        SERVER_KEY = "Server Key";

    function genNonce (len) {
        var nonce, hex, i;

        function toHex (n) {
            if (n < 16) {
                return "0" + n.toString(16);
            }
            return n.toString(16);
        }

        nonce = bitops.NONCE((len || 32) / 2);

        hex = "";
        for (i = 0; i < nonce.length; i++) {
            hex += toHex(nonce[i]);
        }

        return hex;
    }

    function parseChallenge (chal) {
        var dtives, tokens, dtiv, i, len;

        dtives = {};
        tokens = chal.split(/,(?=(?:[^"]|"[^"]*")*$)/);

        for (i = 0, len = tokens.length; i < len; i++) {
            dtiv = (/(\w+)=["]?([^"]+)["]?$/).exec(tokens[i]);
            if (dtiv) {
                dtives[dtiv[1]] = dtiv[2];
            }
        }
        return dtives;
    }

    /* The characters ',' or '=' in usernames are sent as '=2C' and
     * '=3D' respectively. */
    function escapename (name) {
        var escaped, curr, i;

        escaped = [];
        curr = "";
        for (i = 0; i < name.length; i++) {
            curr = name[i];
            if (curr === ",") {
                escaped.push("=2C");
            } else if (curr === "=") {
                escaped.push("=3D");
            } else {
                escaped.push(curr);
            }
        }
        return escaped.join("");
    }

    /* See https://developer.mozilla.org/en-US/docs/Web/API/WindowBase64/btoa */
    function utf8ToBase64 (str) {
        return window.btoa(unescape(encodeURIComponent(str)));
    }

    function compareArrays (arr1, arr2) {
        var i;

        if (!arr1 && !arr2) {
            return true;
        }

        if (!arr1 || !arr2) {
            return false;
        }

        if (arr1.length !== arr2.length) {
            return false;
        }

        for (i = 0; i < arr1.length; ++i) {
            if (arr1[i] !== arr2[i]) {
                return false;
            }
        }

        return true;
    }

    /*
     * SCRAM-SHA-1 `Mechanism` constructor.
     *
     * This class implements the SCRAM-SHA-1 SASL mechanism.
     *
     * References:
     *  - [RFC 5802](http://tools.ietf.org/html/rfc5802)
     */
    function Mechanism (options) {
        var o = options || {};

        this.genNonce = o.genNonce || genNonce;
        this.stage = "initial";
        this.completed = false;
    }

    Mechanism.prototype.name = "SCRAM-SHA-1";
    Mechanism.prototype.clientFirst = true;

    Mechanism.prototype.response = function (cred) {
        return RESP[this.stage](this, cred);
    };

    Mechanism.prototype.challenge = function (chal) {
        var values, verifier;

        values = parseChallenge(chal);

        switch (this.stage) {
        case "challenge":
            if (!values.r || !values.s || !values.i) {
                this.error = "An error occurred during the athentication exchange (invalid server-first-message)";
                return false;
            }

            /* s: the base64-encoded salt used by the
             * server for this user.  It is sent by the server in its first
             * message to the client. */
            this.salt = new Uint8Array(atob(values.s).split("").map(function (c) {
                return c.charCodeAt(0);
            }));

            /* i: iteration count for the selected hash function and user.
             * MUST be sent by the server along with the user's salt. */
            this.iterationCount = parseInt(values.i || "0", 10);

            /* r: a sequence of random printable ASCII characters excluding ','
             * (which forms the nonce used as input to
             * the hash function).  No quoting is applied to this string.  As
             * described earlier, the client supplies an initial value in its
             * first message, and the server augments that value with its own
             * nonce in its first response.  It is important that this value be
             * different for each authentication.
             * The client MUST verify that the initial
             * part of the nonce used in subsequent messages is the same as the
             * nonce it initially specified. */
            this.serverNonce = values.r;
            if (this.serverNonce.slice(0, this.clientNonce.length) !== this.clientNonce) {
                this.error = "An error occurred during the athentication exchange (invalid nonce)";
                return false;
            }

            this.firstChallenge = chal;

            break;
        case "final":
            if (!values.v && !values.e) {
                this.error = "An error occurred during the athentication exchange (invalid server-final-message)";
                return false;
            }

            /* e: specifies an error that occurred during
             * authentication exchange.  It is sent by the server in its final
             * message and can help diagnose the reason for the authentication
             * exchange failure.  On failed authentication, the entire server-
             * final-message is OPTIONAL; specifically, a server implementation
             * MAY conclude the SASL exchange with a failure without sending the
             * server-final-message.  This results in an application-level error
             * response without an extra round-trip.  If the server-final-message
             * is sent on authentication failure, then the "e" attribute MUST be
             * included. */
            if (values.e) {
                this.error = "An error occurred during the athentication exchange (" + values.e + ")";
                return false;
            }
            this.error = values.e;

            /* v: base64-encoded ServerSignature.  It
             * is sent by the server in its final message, and is used by the
             * client to verify that the server has access to the user's
             * authentication information. */
            verifier = new Uint8Array(atob(values.v).split("").map(function (c) {
                return c.charCodeAt(0);
            }));

            if (!compareArrays(verifier, this.serverSignature)) {
                this.error = "An error occurred during the athentication exchange (invalid server signature)";
                return false;
            }

            this.completed = true;

            break;

        default:

            /* Should never happen */
            return false;
        }

        return true;
    };

    RESP.initial = function (mech, cred) {
        var authzid, nonce, username, result;

        if (!cred.username) {
            throw new Error("No username specified.");
        }

        /* The "client-first-message" containing:
         * - a GS2 header consisting of a flag indicating whether channel
         *   binding is supported-but-not-used, not supported, or used, and an
         *   optional SASL authorization identity;
         *
         * - SCRAM username and a random, unique nonce attributes.
         */

        /* client nonce */
        mech.clientNonce = mech.genNonce();

        authzid = "";
        if (cred.authzid) {
            /* This is an optional attribute, it specifies an
             * authorization identity.  A client may include it in its first
             * message to the server if it wants to authenticate as one user, but
             * subsequently act as a different user. */
            authzid = "a=" + escapename(cred.authzid);
        }

        /* We do not support channel binding, then we MUST use
           an "n" gs2-cbind-flag. The last "," is part of the header */
        mech.gs2Header = "n," + authzid + ",";

        nonce = "r=" + mech.clientNonce;

        /* n: specifies the name of the user whose password is
         * used for authentication (a.k.a. "authentication identity"
         * [RFC4422]).  A client MUST include it in its first message to the
         * server.  If the "a" attribute is not specified (which would
         * normally be the case), this username is also the identity that
         * will be associated with the connection subsequent to
         * authentication and authorization. */
        username = "n=" + escapename(cred.username);

        mech.clientFirstMessageBare = username + "," + nonce;
        result = mech.gs2Header + mech.clientFirstMessageBare;

        mech.stage = "challenge";

        return result;
    };

    RESP.challenge = function (mech, cred) {

        var gs2Header, saltedPassword, clientKey, serverKey, storedKey, authMessage, clientSignature,
            clientProof, result;

        /* c: specifies the base64-encoded GS2 header
         * and channel binding data.  It is sent by the client in its second
         * authentication message.  The attribute data consist of:
         *
         * - the GS2 header from the client's first message (recall that the
         *   GS2 header contains a channel binding flag and an optional
         *   authzid).  This header is going to include channel binding type
         *   prefix (see [RFC5056]), if and only if the client is using
         *   channel binding;
         *
         * - followed by the external channel's channel binding data, if and
         *  only if the client is using channel binding.
         *
         * We do not support channel binding
         */
        gs2Header = utf8ToBase64(mech.gs2Header);

        mech.clientFinalMessageWithoutProof = "c=" + gs2Header + ",r=" + mech.serverNonce;

        saltedPassword = bitops.Hi(cred.password || "", mech.salt, mech.iterationCount);
        clientKey = bitops.HMAC(saltedPassword, CLIENT_KEY);
        serverKey = bitops.HMAC(saltedPassword, SERVER_KEY);

        storedKey = bitops.H(clientKey);
        authMessage = mech.clientFirstMessageBare + "," +
                      mech.firstChallenge + "," +
                      mech.clientFinalMessageWithoutProof;
        clientSignature = bitops.HMAC(storedKey, authMessage);

        clientProof = bitops.XOR(clientKey, clientSignature);

        mech.serverSignature = bitops.HMAC(serverKey, authMessage);

        result = mech.clientFinalMessageWithoutProof + ",p=" +
                 btoa(String.fromCharCode.apply(null, clientProof));

        mech.stage = "final";

        return result;
    };

    return Mechanism;
});

/* ex:set ts=4 et: */

